/*
   cvsd - chroot wrapper to run `cvs pserver' more securely.

   cvsd was originally written by Chris Black <cblack@mokey.com>,
   http://cblack.mokey.com/. That was until release 0.6

   cvsd versions up till 0.8b3 were maintained by Philippe Kehl
   <phkehl@gmx.net>, http://guv.ethz.ch/~flip/cvsd/,
   http://www.oinkzwurgl.org/software/cvsd/.

   after that Arthur de Jong <arthur@tiefighter.et.tudelft.nl>
   took up the work and did a complete rewrite,
   http://tiefighter.et.tudelft.nl/~arthur/cvsd.

   Copyright (C) 1999 Chris Black.
   Copyright (C) 2000 Philippe Kehl.
   Copyright (C) 2001 Arthur de Jong.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <stdlib.h>

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif /* not HAVE_GETOPT_H */
#ifndef HAVE_GETOPT_LONG
#include "getopt_long.h"
#endif /* not HAVE_GETOPT_LONG */

#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pwd.h>
#include <grp.h>

#include "config.h"
#include "xmalloc.h"
#include "reslimit.h"


/* the name by which this program was run. */
static char *program_name;


/* the command that will be run */
static char *cvs_cmd=NULL;


/* the arguments that will be passed */
static char **cvs_args;


/* the environment cvs will run in */
static char *cvs_env[]={
  "HOME=/",
  "TMPDIR=/tmp",
  "CVSUMASK=027", /* i think this one's important. how strictly does cvs follow this? */
  NULL
};


/* the umask value cvs will run as */
static mode_t cvs_umask=0027;


/* the configuration file default from Makefile */
char *cvsd_configfile=DEFAULT_CONFIGFILE;


/* the nice value the daemon should run as */
static int cvsd_nice=0;


/* the user id cvs should be run as */
#define NOUID ((gid_t)-1)
static uid_t cvsd_uid=NOUID;


/* the group id cvs should be run as */
#define NOGID ((gid_t)-1)
static gid_t cvsd_gid=NOGID;


/* cvsd root directory */
static char *cvsd_rootjail=NULL;


/* the pid file */
static char *cvsd_pidfile=NULL;


/* the port to be used for listening */
static int cvsd_port=2401;


/* the maximum number of connections to handle */
static int cvsd_maxconnections=0; /* default=0=unlimited */


/* the number of children spawned */
static int cvsd_numchildren=0;


/* the server socket */
static int cvsd_serversocket=-1;

/* the number of seconds to sleep when no more
   connections can be listened for */
#define SLEEPSECS 5


/* maximum linelength in the configuration file */
#define MAX_LINE_LENGTH 1024


/* display version information */
static void display_version(FILE *fp)
{
  fprintf(fp,"%s %s\n",PACKAGE,VERSION);
  fprintf(fp,"Written by Chris Black, Philippe Kehl and Arthur de Jong.\n\n");
  fprintf(fp,"Copyright (C) 1999,2000,2001 Chris Black, Philippe Kehl and Arthur de Jong.\n"
             "This is free software; see the source for copying conditions.  There is NO\n"
             "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
}


/* display usage information to stdout and exit(status) */
static void display_usage(FILE *fp)
{
  fprintf(fp,"Usage: %s [OPTION]...\n", program_name);
  fprintf(fp,"chroot wrapper to run `cvs pserver' more securely.\n");
  fprintf(fp,"  -f, --config=FILE  use FILE as configfile (default %s)\n",DEFAULT_CONFIGFILE);
  fprintf(fp,"      --help         display this help and exit\n");
  fprintf(fp,"      --version      output version information and exit\n");
  fprintf(fp,"\n"
             "Report bugs to <arthur@tiefighter.et.tudelft.nl>.\n");
}


/* the definition of options for getopt(). see getopt(2) */
static struct option const cvsd_options[] =
{
  { "config",      required_argument, NULL, 'f' },
  { "help",        no_argument,       NULL, 'h' },
  { "version",     no_argument,       NULL, 'V' },
  { NULL, 0, NULL, 0 }
};
#define CVSD_OPTIONSTRING "f:hV"


/* add a string as a command line */
static void add_cvsarg(char *arg)
{
  static int sz=0; /* size of array */
  static int cn=0; /* numver of elements */
  char **tmp;
  int nsz;
  /* first ensure array is large enough */
  if ((cn+2)>sz)
  {
    if (sz==0)
      nsz=10;
    else
      nsz=sz*2;
    tmp=(char **)xmalloc(nsz*sizeof(char *));
    memcpy(tmp,cvs_args,sz*sizeof(char *));
    cvs_args=tmp;
    sz=nsz;
  }
  /* add string to end of array */
  cvs_args[cn++]=arg;
  cvs_args[cn]=NULL;
}


/* parse command line options and set global vars */
static void parse_cmdline(int argc, char *argv[])
{
  int optc;
  while ((optc=getopt_long(argc,argv,CVSD_OPTIONSTRING,cvsd_options,NULL))!=-1)
  {
    switch (optc)
    {
    case 'f': /* -f, --config=FILE  use FILE as configfile */
      cvsd_configfile=(char *)xmalloc((strlen(optarg)+1)*sizeof(char));
      strcpy(cvsd_configfile,optarg);
      break;
    case 'h': /*     --help         display this help and exit */
      display_usage(stdout);
      exit(0);
    case 'V': /*     --version      output version information and exit */
      display_version(stdout);
      exit(0);
    case ':': /* missing required parameter */
    case '?': /* unknown option character or extraneous parameter */
    default:
      fprintf(stderr,"Try `%s --help' for more information.\n",
              program_name);
      exit(1);
    }
  }
  /* check for remaining arguments */
  if (optind<argc)
  {
    fprintf(stderr,"%s: unrecognized option `%s'\n",
            program_name,argv[optind]);
    fprintf(stderr,"Try `%s --help' for more information.\n",
            program_name);
    exit(1);
  }
}


/* find out the uid this string describes
   returns 1 on success, 0 on failure */
static int get_uid(char *str,uid_t *val)
{
  struct passwd *pwent;
  char *tmp;
  /* check if it is a valid numerical uid */
  *val=(uid_t)strtol(str,&tmp,0);
  if ((*str!='\0')&&(*tmp=='\0'))
    return 1;
  /* find by name */
  pwent=getpwnam(str);
  if (pwent!=NULL)
  {
    *val=pwent->pw_uid;
    return 1;
  }
  return 0;
}


/* find the group id this string represents
   returns 1 on success, 0 on failure */
static int get_gid(char *str,gid_t *val)
{
  struct group *grent;
  char *tmp;
  /* check if it is a valid numerical uid */
  *val=(gid_t)strtol(str,&tmp,0);
  if ((*str!='\0')&&(*tmp=='\0'))
    return 1;
  /* find by name */
  grent=getgrnam(str);
  if (grent!=NULL)
  {
    *val=grent->gr_gid;
    return 1;
  }
  return 0;
}


/* read the configuration file */
static void read_configfile(char *filename)
{
  FILE *fp;
  int lnr=0;
  char line[MAX_LINE_LENGTH];
  char *tmp;  /* for manipulating the read line */
  char *tmp2; /* for building cvs command args */
  int i;
  int numrepos=0;
#ifdef HAVE_SETRLIMIT
  char *tmp3; /* TODO: really improve this */
  rlim_t resval;
#endif /* HAVE_SETRLIMIT */
  /* open configuration file */
  if ((fp=fopen(filename,"r"))==NULL)
  {
    syslog(LOG_ERR,"cannot open config file (%s): %s",filename,strerror(errno));
    exit(1);
  }
#ifdef DEBUG
  syslog(LOG_DEBUG,"reading config file (%s)",filename);
#endif /* DEBUG */
  /* read all lines in the file */
  while (fgets(line,MAX_LINE_LENGTH,fp)!=NULL)
  {
    lnr++;
    /* strip newline */
    i=(int)strlen(line);
    if ((i>0)&&(line[i-1]=='\n'))
    {
      line[i-1]='\0';
    }
    else
    {
      syslog(LOG_ERR,"%s:%d: line too long or last line missing newline",filename,lnr);
      exit(1);
    }
    /* strip out comments */
    tmp=strchr(line,'#');
    if (tmp!=NULL) *tmp='\0';
    /* convert tabs to spaces */
    tmp=line;
    while (*tmp!='\0')
    {
      if (*tmp=='\t') *tmp=' ';
      tmp++;
    }
    /* remove trailing spaces */
    tmp--;
    while ((tmp>=line)&&(*tmp==' '))
    {
      *tmp='\0';
      tmp--;
    }
    /* skip beginning spaces */
    tmp=line;
    while (*tmp==' ')
      tmp++;
    /* find identifier */
    if (strncmp(tmp,"RootJail ",9)==0)
    {
      /* skip space */
      for (tmp+=9;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: RootJail: argument missing",filename,lnr);
        exit(1);
      }
      /* only accept absolute paths. abort otherwise.*/
      if ((*tmp!='/')&&(strcmp(tmp,"none")!=0))
      {
        syslog(LOG_ERR,"%s:%d: RootJail: '%s' must be an absolute path or 'none'",filename,lnr,tmp);
        exit(1);
      }
      /* save value */
      if (cvsd_rootjail!=NULL)
      {
        syslog(LOG_ERR,"%s:%d: RootJail: can only be specified once",filename,lnr);
      }
      else
      {
        cvsd_rootjail=(char *)xmalloc((strlen(tmp)+1)*sizeof(char));
        strcpy(cvsd_rootjail,tmp);
      }
    }
    else if (strncmp(tmp,"Nice ",5)==0)
    {
      /* skip space */
      for (tmp+=5;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: Nice: argument missing",filename,lnr);
        exit(1);
      }
      /* check if it is a valid numerical uid */
      cvsd_nice=(int)strtol(tmp,&tmp2,0);
      if (*tmp2!='\0')
      {
        syslog(LOG_ERR,"%s:%d: Nice: wrong argument '%s'",filename,lnr,tmp);
        exit(1);
      }
    }
    else if (strncmp(tmp,"Uid ",4)==0)
    {
      /* skip space */
      for (tmp+=4;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: Uid: argument missing",filename,lnr);
        exit(1);
      }
      if (!get_uid(tmp,&cvsd_uid))
      {
        syslog(LOG_ERR,"%s:%d: Uid: wrong argument '%s'",filename,lnr,tmp);
        exit(1);
      }
    }
    else if (strncmp(tmp,"Gid ",4)==0)
    {
      /* skip space */
      for (tmp+=4;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: Gid: argument missing",filename,lnr);
        exit(1);
      }
      if (!get_gid(tmp,&cvsd_gid))
      {
        syslog(LOG_ERR,"%s:%d: Gid: wrong argument '%s'",filename,lnr,tmp);
        exit(1);
      }
    }
    else if (strncmp(tmp,"Port ",5)==0)
    {
      /* skip space */
      for (tmp+=5;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: Port: argument missing",filename,lnr);
        exit(1);
      }
      cvsd_port=(int)strtol(tmp,&tmp2,0);
      if (*tmp2!='\0')
      {
        syslog(LOG_ERR,"%s:%d: Port: wrong argument '%s'",filename,lnr,tmp);
        exit(1);
      }
    }
    else if (strncmp(tmp,"MaxConnections ",15)==0)
    {
      /* skip space */
      for (tmp+=15;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: MaxConnections: argument missing",filename,lnr);
        exit(1);
      }
      cvsd_maxconnections=(int)strtol(tmp,&tmp,0);
      if (*tmp!='\0')
      {
        syslog(LOG_ERR,"%s:%d: MaxConnections: wrong argument '%s'",filename,lnr,tmp);
        exit(1);
      }
    }
    else if (strncmp(tmp,"Repos ",6)==0)
    {
      /* skip space */
      for (tmp+=6;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: Repos: argument missing",filename,lnr);
        exit(1);
      }
      /* only accept absolute paths. abort otherwise.*/
      if (*tmp!='/')
      {
        syslog(LOG_ERR,"%s:%d: Repos: '%s' must be an absolute path",filename,lnr,tmp);
        exit(1);
      }
      /* add --allow-root=path to cvs parameters */
      tmp2=(char *)xmalloc((13+strlen(tmp)+1)*sizeof(char));
      strcpy(tmp2,"--allow-root=");
      strcat(tmp2,tmp);
      add_cvsarg(tmp2);
      numrepos++;
    }
    else if (strncmp(tmp,"PidFile ",8)==0)
    {
      /* skip space */
      for (tmp+=8;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: PidFile: argument missing",filename,lnr);
        exit(1);
      }
      if (cvsd_pidfile==NULL)
      {
        cvsd_pidfile=(char *)xmalloc((strlen(tmp)+1)*sizeof(char));
        strcpy(cvsd_pidfile,tmp);
      }
    }
    else if (strncmp(tmp,"CvsCommand ",11)==0)
    {
      /* skip space */
      for (tmp+=11;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: CvsCommand: argument missing",filename,lnr);
        exit(1);
      }
      /* only accept absolute paths, abort otherwise */
      if (*tmp!='/')
      {
        syslog(LOG_ERR,"%s:%d: CvsCommand: '%s' must be an absolute path",filename,lnr,tmp);
        exit(1);
      }
      /* save value */
      if (cvs_cmd!=NULL)
      {
        syslog(LOG_ERR,"%s:%d: CvsCommand: can only be specified once",filename,lnr);
      }
      else
      {
        cvs_cmd=(char *)xmalloc((strlen(tmp)+1)*sizeof(char));
        strcpy(cvs_cmd,tmp);
      }
    }
    else if (strncmp(tmp,"Limit ",6)==0)
    {
#ifdef HAVE_SETRLIMIT
      /* skip space */
      for (tmp+=6;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: Limit: resourcename missing",filename,lnr);
        exit(1);
      }
      /* find resource name */
      tmp3=tmp;
      for (;(*tmp!=' ')&&(*tmp!='\0');tmp++);
      for (;*tmp==' ';tmp++) *tmp='\0';
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: Limit: limit value missing",filename,lnr);
        exit(1);
      }
      /* read value */
      if ( (strcmp(tmp,"none")==0) ||
           (strcmp(tmp,"inf")==0) ||
           (strcmp(tmp,"unlimited")==0) )
        resval=RLIM_INFINITY;
      else
        resval=0;
      switch (getresourcetype(tmp3))
      {
      case RESTYPE_NUM:  /* numeric value */
      case RESTYPE_SIZE: /* size value */
        if (resval==RLIM_INFINITY) break;
        resval=(rlim_t)strtol(tmp,&tmp2,0);
        if ((tmp2[0]=='k')&&(tmp2[1]=='\0'))
          resval*=(rlim_t)1024;
        else if ((tmp2[0]=='m')&&(tmp2[1]=='\0'))
          resval*=(rlim_t)(1024*1024);
        else if ((tmp2[0]!='\0')&&(tmp2[0]!='b'))
        {
          syslog(LOG_ERR,"%s:%d: Limit: wrong value '%s'",filename,lnr,tmp);
          exit(1);
        }
        break;
      case RESTYPE_TIME: /* time value */
        if (resval==RLIM_INFINITY) break;
        resval=(rlim_t)strtol(tmp,&tmp2,0);
        if ((tmp2[0]=='m')&&(tmp2[1]=='\0'))
          resval*=(rlim_t)60;
        else if ((tmp2[0]==':')&&(tmp2[1]!='\0'))
        {
          tmp2++;
          resval*=(rlim_t)60;
          resval+=(rlim_t)strtol(tmp2,&tmp2,0);
          if (tmp2[0]!='\0')
          {
            syslog(LOG_ERR,"%s:%d: Limit: wrong value '%s'",filename,lnr,tmp);
            exit(1);
          }
        }
        else if ((tmp2[0]!='\0')&&(tmp2[0]!='s'))
        {
          syslog(LOG_ERR,"%s:%d: Limit: wrong value '%s'",filename,lnr,tmp);
          exit(1);
        }
        break;
      default:
        syslog(LOG_ERR,"%s:%d: Limit: unknown or unsupported resource '%s'",filename,lnr,tmp2);
        exit(1);
      }
      /* save the limit, execute with setreslimits() */
      savereslimit(tmp3,resval,resval);
#else /* HAVE_SETRLIMIT */
      syslog(LOG_ERR,"%s:%d: Limit: not supported on this system",filename,lnr);
      exit(1);
#endif /* not HAVE_SETRLIMIT */
    }
    else if (*tmp!='\0') /* line should be empty */
    {
      syslog(LOG_ERR,"%s:%d: unrecognized line",filename,lnr);
      exit(1);
    }
  }
  fclose(fp);
  if (numrepos==0)
  {
    syslog(LOG_ERR,"%s: contains no 'Repos' statements",filename);
    exit(1);
  }
}


/* get a name of a signal with a given signal number */
static const char *signame(int signum)
{
  switch (signum)
  {
  case SIGHUP:  return "SIGHUP";  /* Hangup detected */
  case SIGINT:  return "SIGINT";  /* Interrupt from keyboard */
  case SIGQUIT: return "SIGQUIT"; /* Quit from keyboard */
  case SIGILL:  return "SIGILL";  /* Illegal Instruction */
  case SIGABRT: return "SIGABRT"; /* Abort signal from abort(3) */
  case SIGFPE:  return "SIGFPE";  /* Floating point exception */
  case SIGKILL: return "SIGKILL"; /* Kill signal */
  case SIGSEGV: return "SIGSEGV"; /* Invalid memory reference */
  case SIGPIPE: return "SIGPIPE"; /* Broken pipe */
  case SIGALRM: return "SIGALRM"; /* Timer signal from alarm(2) */
  case SIGTERM: return "SIGTERM"; /* Termination signal */
  case SIGUSR1: return "SIGUSR1"; /* User-defined signal 1 */
  case SIGUSR2: return "SIGUSR2"; /* User-defined signal 2 */
  case SIGCHLD: return "SIGCHLD"; /* Child stopped or terminated */
  case SIGCONT: return "SIGCONT"; /* Continue if stopped */
  case SIGSTOP: return "SIGSTOP"; /* Stop process */
  case SIGTSTP: return "SIGTSTP"; /* Stop typed at tty */
  case SIGTTIN: return "SIGTTIN"; /* tty input for background process */
  case SIGTTOU: return "SIGTTOU"; /* tty output for background process */
#ifdef SIGBUS
  case SIGBUS:  return "SIGBUS";  /* Bus error */
#endif
#ifdef SIGPOLL
  case SIGPOLL: return "SIGPOLL"; /* Pollable event */
#endif
#ifdef SIGPROF
  case SIGPROF: return "SIGPROF"; /* Profiling timer expired */
#endif
#ifdef SIGSYS
  case SIGSYS:  return "SIGSYS";  /* Bad argument to routine */
#endif
#ifdef SIGTRAP
  case SIGTRAP: return "SIGTRAP"; /* Trace/breakpoint trap */
#endif
#ifdef SIGURG
  case SIGURG:  return "SIGURG";  /* Urgent condition on socket */
#endif
#ifdef SIGVTALRM
  case SIGVTALRM: return "SIGVTALRM"; /* Virtual alarm clock */
#endif
#ifdef SIGXCPU
  case SIGXCPU: return "SIGXCPU"; /* CPU time limit exceeded */
#endif
#ifdef SIGXFSZ
  case SIGXFSZ: return "SIGXFSZ"; /* File size limit exceeded */
#endif
  default:      return "UNKNOWN";
  }
}


/* signal handler for dieing children */
static RETSIGTYPE sigchld_handler(int signum)
{
  int status;
  pid_t cpid;
  /* TODO?: replace syslog with something else? */
  /* loop until all pending child processes have finished
     (kill all the zombies) */
  do
  {
    cpid=waitpid(-1,&status,WNOHANG);
    if (cpid>0) /* if there was a terminating child */
    {
      /* shouldn't have a race here because we don't have threads */
      cvsd_numchildren--;
      if (WIFSIGNALED(status))
      {
        syslog(LOG_INFO,"cvs command exited on signal %s (%d) with exit-status %d\n",
                        signame((int)WTERMSIG(status)),
                        (int)WTERMSIG(status),
                        (int)WEXITSTATUS(status));
      }
      else
      {
        syslog(LOG_INFO,"cvs command exited%s with exit-status %d\n",
                        WIFEXITED(status)?"":" abnormally",
                        (int)WEXITSTATUS(status));
      }
    }
  }
  while (cpid>0);
}


/* signal handler for closing down */
static void sigexit_handler(int signum)
{
  syslog(LOG_INFO,"caught signal %s (%d), shutting down",
                  signame(signum),signum);
  if (cvsd_serversocket>=0)
    close(cvsd_serversocket);
  exit(1);
}


/* return a socket for doing accept() on */
static int getlistensocket(int port)
{
  int sock;
  struct sockaddr_in addr;
  int flag;

  /* create a socket for tcp connections */
  if ((sock=socket(AF_INET,SOCK_STREAM,0))<0)
  {
    syslog(LOG_ERR,"cannot create socket: %s",strerror(errno));
    exit(1);
  }
  
  /* set some socket options */
  flag=1;
  if (setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,&flag,sizeof(int))!=0)
  {
    syslog(LOG_ERR,"setsockopt(SO_REUSEADDR) failed: %s",strerror(errno));
    exit(1);
  }

  /* bind the socket to the specified port */
  addr.sin_family=AF_INET;
  addr.sin_port=htons(port);
  addr.sin_addr.s_addr=INADDR_ANY;
  if (bind(sock,(struct sockaddr *)&addr,(socklen_t)sizeof(addr))<0)
  {
    syslog(LOG_ERR,"cannot bind to port %d: %s",port,strerror(errno));
    exit(1);
  }

  /* start listening for connections */
  if (listen(sock,1)<0)
  {
    syslog(LOG_ERR,"cannot listen on socket: %s",strerror(errno));
    exit(1);
  }
  return sock;
}


/* accept a connection on the socket and spawn a child cvs */
static void acceptconnection(int sock)
{
  int csock;
  struct sockaddr_in caddr;
  socklen_t clen;
  pid_t cpid;

  /* accept a new connection */
  clen=(socklen_t)sizeof(caddr);
  if ((csock=accept(sock,(struct sockaddr *)&caddr,&clen))<0)
  {
    syslog(LOG_ERR,"accept() failed: %s",strerror(errno));
    return;
  }
  if (clen!=(socklen_t)sizeof(caddr))
  {
    syslog(LOG_ERR,"accept() parameter error");
    return;
  }

  /* we have a possible race here? (probably atomic) */
  cvsd_numchildren++;

  /* we now have a connection from inet_ntoa(caddr.sin_addr)
     port ntohs(caddr.sin_port) */
  syslog(LOG_INFO,"connection from %s:%d",
                  inet_ntoa(caddr.sin_addr),
                  (int)ntohs(caddr.sin_port));

  /* fork off a child to handle the connection */
  switch (cpid=fork())
  {
  case 0: /* we are the child */
    /* close server socket */
    close(sock);
    /* connect socket to stdin/stdout/stderr */
    if ( (dup2(csock,0)<0) || /* stdin */
         (dup2(csock,1)<0) || /* stdout */
         (dup2(csock,2)<0) )  /* stderr */
    {
      syslog(LOG_ERR,"dup failed: %s",strerror(errno));
      exit(1);
    }
    /* TODO: ?? close all other file descriptors (syslog etc?) */
    /* set limits */
    setreslimits();
    /* execute cvs */
    execve(cvs_cmd,cvs_args,cvs_env);
    /* if we are here there has been an error */
    syslog(LOG_ERR,"error executing `cvs': %s",strerror(errno));
    exit(1);
    break;
  case -1: /* we are the parent, but have an error */
    syslog(LOG_ERR,"fork() failed: %s",strerror(errno));
    close(sock);
    exit(1);
    break;
  }

  /* we are the parent and have a cpid */
#ifdef DEBUG
  syslog(LOG_DEBUG,"fork() succeeded (child pid=%d)",(int)cpid);
#endif

  /* close the socket to prevent confusion */
  close(csock);
}


/* fork and return as child */
static void daemonize(void)
{
  /* TODO: use daemon(0,0) if available, otherwise this code */
  switch (fork())
  {
  case 0: /* we are the child */
    break;
  case -1: /* we are the parent, but have an error */
    fprintf(stderr,"%s: unable to daemonize: %s\n",
                   program_name,strerror(errno));
    exit(1);
    break;
  default: /* we are the parent and we're done*/
    exit(0);
  }
  /* TODO: redirect file descriptors to /dev/null */
}


/* write the current process id to the specified file */
static void write_pidfile(char *filename)
{
  FILE *fp;
  if (filename!=NULL)
  {
    if ((fp=fopen(filename,"w"))==NULL)
    {
      syslog(LOG_ERR,"cannot open pid file (%s): %s",filename,strerror(errno));
      exit(1);
    }
    fprintf(fp,"%d\n",(int)getpid());
    fclose(fp);
  }
}


/* try to install signal handler and check result */
static void install_sighandler(int signum,void (*handler) (int))
{
  if (signal(signum,handler)==SIG_ERR) 
  {
    syslog(LOG_ERR,"error installing signal handler for '%s'",signame(signum));
    if (cvsd_serversocket>=0)
      close(cvsd_serversocket);
    exit(1);
  }
}


/* the main program... */
int main(int argc,char *argv[])
{
#ifdef DEBUG
  int i,j;
  char *cvs_cmdline=NULL;
#endif /* DEBUG */

  /* save program name */
  program_name=argv[0];

  /* parse the command line */
  parse_cmdline(argc,argv);

  /* daemonize */
  daemonize();

  /* open system logger...
     (from here only use syslog) */
  openlog("cvsd",LOG_PID,LOG_DAEMON);
  syslog(LOG_INFO,"version %s starting",VERSION);

  /* begin building command line */
  add_cvsarg("cvs");
  add_cvsarg("-f"); /* do not read ~/.cvsrc */
#ifdef DEBUG
  add_cvsarg("-t"); /* verbose */
#endif /* DEBUG */

  /* read the configuration file */
  read_configfile(cvsd_configfile);
  
  /* specify cvs_command if not set earlier */
  if (cvs_cmd==NULL)
  {
    cvs_cmd=( (cvsd_rootjail!=NULL) &&
              (strcmp(cvsd_rootjail,"none")!=0) )?"/bin/cvs":CVS_LOCATION;
  }

  /* add cvs action */
  add_cvsarg("pserver"); /* run cvs as pserver */

  /* write pidfile */
  write_pidfile(cvsd_pidfile);

  /* set the umask */
  umask(cvs_umask);

  /* bind port */
  cvsd_serversocket=getlistensocket(cvsd_port);
  
  /* do the chroot stuff */
  if ( (cvsd_rootjail!=NULL) &&
       (strcmp(cvsd_rootjail,"none")!=0) )
  {
    if (chdir(cvsd_rootjail)!=0)
    {
      syslog(LOG_ERR,"cannot chdir(%s): %s",cvsd_rootjail,strerror(errno));
      close(cvsd_serversocket);
      exit(1);
    }
    if (chroot(cvsd_rootjail)!=0)
    {
      syslog(LOG_ERR,"cannot chroot(%s): %s",cvsd_rootjail,strerror(errno));
      close(cvsd_serversocket);
      exit(1);
    }
  }
  /* just to be sure */
  chdir("/");
  
  /* renice */
  errno=0; /* for strange nice implementations values */
  nice(cvsd_nice);
  if (errno!=0)
  {
    syslog(LOG_ERR,"cannot nice(%d): %s",(int)cvsd_nice,strerror(errno));
    close(cvsd_serversocket);
    exit(1);
  }

  /* change to cvs gid */
  if ((cvsd_gid!=NOGID)&&(setgid(cvsd_gid))!=0)
  {
    syslog(LOG_ERR,"cannot setgid(%d): %s",(int)cvsd_gid,strerror(errno));
    close(cvsd_serversocket);
    exit(1);
  }

  /* change to cvs uid */
  if ((cvsd_uid!=NOUID)&&(setuid(cvsd_uid)!=0))
  {
    syslog(LOG_ERR,"cannot setuid(%d): %s",(int)cvsd_uid,strerror(errno));
    close(cvsd_serversocket);
    exit(1);
  }

  /* build cvs command line */
#ifdef DEBUG
  j=strlen(cvs_cmd)+1;
  for (i=1;cvs_args[i]!=NULL;i++)
    j+=strlen(cvs_args[i])+1;
  cvs_cmdline=(char *)xmalloc(j*sizeof(char));
  strcpy(cvs_cmdline,cvs_cmd);
  for (i=1;cvs_args[i]!=NULL;i++)
  {
    strcat(cvs_cmdline," ");
    strcat(cvs_cmdline,cvs_args[i]);
  }
  syslog(LOG_DEBUG,"cvs command to execute: '%s'",cvs_cmdline);
#endif /* DEBUG */

  /* install signalhandler for terminating children */
  install_sighandler(SIGCHLD,sigchld_handler);
  
  /* install signalhandlers for some other signals */
  install_sighandler(SIGHUP, sigexit_handler);
  install_sighandler(SIGINT, sigexit_handler);
  install_sighandler(SIGQUIT,sigexit_handler);
  install_sighandler(SIGILL, sigexit_handler);
  install_sighandler(SIGABRT,sigexit_handler);
  install_sighandler(SIGSEGV,sigexit_handler);
  install_sighandler(SIGPIPE,sigexit_handler);
  install_sighandler(SIGALRM,sigexit_handler);
  install_sighandler(SIGTERM,sigexit_handler);
  install_sighandler(SIGUSR1,sigexit_handler);
  install_sighandler(SIGUSR2,sigexit_handler);

  syslog(LOG_INFO,"listening for connections");

  /* start waiting for incoming connections */
  while (1)
  {
    /* if we are ready to handle a connection do so */
    if ((cvsd_maxconnections>0)&&(cvsd_numchildren<cvsd_maxconnections))
    {
      /* wait for a new connection */
      acceptconnection(cvsd_serversocket);
    }
    else /* not ready for connection */
    {
      /* log that the connections are full */
      syslog(LOG_DEBUG,"too many connections (%d), listening suspended",cvsd_numchildren);
      /* wait until children have died */
      while (cvsd_numchildren>=cvsd_maxconnections)
      {
        /* sleep unblocks on a signal */
        sleep(SLEEPSECS);
      }
      /* log that the connections are full */
      syslog(LOG_DEBUG,"resuming listening for connections");
    }
  }

  return 1; /* keep compiler happy */
}
